/***********************************************************
 * $Id: ProxyFactory.java 402 2011-01-01 09:31:17Z util $
 * 
 * AOP Utility classes of the clazzes.org project
 * http://www.clazzes.org
 *
 * Created: 2010-08-31
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 ***********************************************************/

package org.clazzes.util.aop;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.HashSet;
import java.util.Set;
import java.util.regex.Pattern;

import org.aopalliance.intercept.MethodInterceptor;
import org.aopalliance.intercept.MethodInvocation;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * <p>A lightweight proxy factory for instantiating intercepted instances
 * of interface implementations.</p>
 * 
 * <p>If you need special sets of interceptors and multiple implementing interfaces, use
 * {@link #getProxy()} for creating proxies.</p>
 * 
 * <p>For most cases you might it might be sufficient to use {@link #getTypedProxy(Object, Class)},
 * which allows you to use a singleton {@link ProxyFactory} for generating multiple proxies.
 */
public class ProxyFactory
{
    private static final Logger log = LoggerFactory.getLogger(ProxyFactory.class);
    
    private Object target;
    private ClassLoader classLoader;
    @SuppressWarnings("rawtypes")
    private Class[] interfaces;
    private MethodInterceptor[] interceptors;
    private Pattern includeMethods;
    private Pattern excludeMethods;
    
    private final class ProxyInterceptionInvocationHandler implements InvocationHandler {

        private final InvocationHandler handler;
        private final Set<Method> intereptedMethods;

        /**
         * @param handler The invocation handler to invoke in the end.
         */
        public ProxyInterceptionInvocationHandler(InvocationHandler handler, Set<Method> intereptedMethods)
        {
            super();
            this.handler = handler;
            this.intereptedMethods = intereptedMethods;
        }

        public Object invoke(Object proxy, Method method, Object[] args) throws Throwable
        {
            if (!this.intereptedMethods.contains(method)) {
                return this.handler.invoke(proxy, method, args);
            }
            
            MethodInvocation invocation =
                new InvocationHandlerMethodInvocation(this.handler,proxy,method,args);
            
            int i=ProxyFactory.this.interceptors.length;
            
            while (--i > 0) {
                invocation = new InterceptedMethodInvocation(invocation,ProxyFactory.this.interceptors[i]);
            }
 
            return ProxyFactory.this.interceptors[0].invoke(invocation);
        }
    }
    
    private final class ReflectiveInterceptionInvocationHandler implements InvocationHandler {

        private final Object target;
        private final Set<Method> intereptedMethods;

        /**
         * @param handler The invocation handler to invoke in the end.
         */
        public ReflectiveInterceptionInvocationHandler(Object target, Set<Method> intereptedMethods)
        {
            super();
            this.target = target;
            this.intereptedMethods = intereptedMethods;
        }

        public Object invoke(Object proxy, Method method, Object[] args) throws Throwable
        {
            MethodInvocation invocation =
                new ReflectiveMethodInvocation(this.target,method,args);
            
            if (!this.intereptedMethods.contains(method)) {
                return invocation.proceed();
            }

            int i=ProxyFactory.this.interceptors.length;
            
            while (--i > 0) {
                invocation = new InterceptedMethodInvocation(invocation,ProxyFactory.this.interceptors[i]);
            }
 
            return ProxyFactory.this.interceptors[0].invoke(invocation);
        }
    }
    
    @SuppressWarnings("rawtypes")
    private Set<Method> makeIntereptedMethods(Class[] ifaces) {
        
        Set<Method> ret = new HashSet<Method>();
        
        for (Class iface : ifaces) {
            
            for (Method m : iface.getMethods()) {

                String methodName = m.getName();

                if (this.excludeMethods != null && this.excludeMethods.matcher(methodName).matches()) {
                    if (log.isDebugEnabled())
                        log.debug(" - omitting excluded method ["+m+"].");
                    continue;
                }
                if (this.includeMethods != null && !this.includeMethods.matcher(methodName).matches()) {
                    if (log.isDebugEnabled())
                        log.debug(" - omitting not included method ["+m+"].");
                    continue;
                }
                if (log.isDebugEnabled())
                    log.debug(" + registering method ["+m+"].");

                ret.add(m);
            }
        }
        
        return ret;
    }
    
    /**
     * <p>This method facilitates the use of a proxy singleton to generate more proxies for
     * different targets with changing interfaces but with the same set of interceptors.
     * If you need more than one implemented interface or your interceptors are different
     * for each target, use separate {@link ProxyFactory} instances with different bean
     * configurations.</p>
     * 
     * <p>This method ignores the <code>target</code> and <code>interfaces</code>
     * bean properties.</p> 
     * 
     * @param _target The target to wrap.
     * @param iface The interface of the returned proxy.
     * @return A new JDK proxy which implements the given interface and involves the
     *         specified interceptors.
     */
    public <T> T getTypedProxy(Object _target, Class<T> iface) {
        
        Class<T>[] ifaces = new Class[] { iface };
        
        return (T)getProxyInternal(_target,ifaces);
    }
    
    /**
     * @return a new proxy, which implements the given interfaces and involves the
     *         specified interceptors.
     *         
     */
    @SuppressWarnings({"rawtypes" })
    public Object getProxy() {
        
        Class[] ifaces;
        
        if (this.interfaces == null)
            ifaces = this.target.getClass().getInterfaces();
        else
            ifaces = this.interfaces;
        
        return getProxyInternal(this.target,ifaces);
    }
    
    @SuppressWarnings("rawtypes")
    protected Object getProxyInternal(Object _target, Class[] ifaces) {
        
        Set<Method> interceptedMethods = makeIntereptedMethods(ifaces); 
        
        InvocationHandler h;
        
        if (_target != null && Proxy.isProxyClass(_target.getClass())) {
            h = new ProxyInterceptionInvocationHandler(Proxy.getInvocationHandler(_target),interceptedMethods);
        }
        else {
            h = new ReflectiveInterceptionInvocationHandler(_target,interceptedMethods);
        }
        
        ClassLoader cl;
        
        // rationale:
        //
        // If the class loader is not set, the target is the class, which is closest to
        // the resulting proxy. Normally, target is an implementation class, which lives
        // inside the deployment's container realm. In OSGi contexts, the target may well
        // be a service proxy, which also lives inside the current bundle's context. 
        if (this.classLoader == null) {
            if (_target != null) {
                cl = _target.getClass().getClassLoader();
            }
            else {
                cl = Thread.currentThread().getContextClassLoader();
            }
        } else {
            cl = this.classLoader;
        }
        
        return Proxy.newProxyInstance(cl,ifaces,h);
    }

    /**
     * @return The invocation target.
     */
    public Object getTarget()
    {
        return this.target;
    }

    /**
     * @param target the target to set
     */
    public void setTarget(Object target)
    {
        this.target = target;
    }

    /**
     * @return The class loader to set. If <code>null</code>, the class loader of
     *         the target object is used.
     */
    public ClassLoader getClassLoader()
    {
        return this.classLoader;
    }

    /**
     * @param classLoader the classLoader to set
     */
    public void setClassLoader(ClassLoader classLoader)
    {
        this.classLoader = classLoader;
    }

    /**
     * @return the interfaces
     */
    @SuppressWarnings("rawtypes")
    public Class[] getInterfaces()
    {
        return this.interfaces;
    }

    /**
     * @param interfaces the interfaces to set
     */
    @SuppressWarnings("rawtypes")
    public void setInterfaces(Class[] interfaces)
    {
        this.interfaces = interfaces;
    }

    /**
     * Convenience setter, which sets the implemented interfaces to a single interface.
     * 
     * @param iface the interface to set
     */
    @SuppressWarnings("rawtypes")
    public void setInterface(Class iface)
    {
        this.interfaces = new Class[] {iface};
    }

    /**
     * @return the interceptors invoked in the given order. The first interceptor is the outermost
     *         one.
     */
    public MethodInterceptor[] getInterceptors()
    {
        return this.interceptors;
    }

    /**
     * @param interceptors the interceptors to set
     */
    public void setInterceptors(MethodInterceptor[] interceptors)
    {
        this.interceptors = interceptors;
    }

    /**
     * Convenience setter, which sets the invoked interceptors to a single instance.
     * 
     * @param interceptor the interceptor to set
     */
    public void setInterceptor(MethodInterceptor interceptor)
    {
        this.interceptors = new MethodInterceptor[] {interceptor};
    }

    /**
     * @return An optional regular expression for selecting methods, which are
     *         intercepted. If <code>null</code>, all interface methods are included. 
     */
    public Pattern getIncludeMethods()
    {
        return this.includeMethods;
    }

    /**
     * @param includeMethods the includeMethods to set
     */
    public void setIncludeMethods(Pattern includeMethods)
    {
        this.includeMethods = includeMethods;
    }

    /**
     * @return An optional regular expression for omitting methods, which are
     *         intercepted. If <code>null</code>, no interface method is omitted. 
     */
    public Pattern getExcludeMethods()
    {
        return this.excludeMethods;
    }

    /**
     * @param excludeMethods the excludeMethods to set
     */
    public void setExcludeMethods(Pattern excludeMethods)
    {
        this.excludeMethods = excludeMethods;
    }
    
}
