/***********************************************************
 * $Id: ThreadLocalManager.java 402 2011-01-01 09:31:17Z util $
 * 
 * AOP Utility classes of the clazzes.org project
 * http://www.clazzes.org
 *
 * Created: 31.08.2010
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 ***********************************************************/

package org.clazzes.util.aop;

import java.util.HashMap;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * This class provides for binding resources like pooled
 * database or remoting clients to the current thread.
 */
public class ThreadLocalManager
{
    private static final Logger log = LoggerFactory.getLogger(ThreadLocalManager.class);
    
    private static final ThreadLocal<Map<String,Object>> threadLocal =
        new ThreadLocal<Map<String,Object>>();
    
    
    /**
     * Bind a resource from a singleton factory to the current thread.
     * 
     * @param key A key for storing the given resource, usually a unique
     *            string like a JDBC URL or a remoting connection address.
     * @param resource A resource like a JDBC connection or a remoting context. 
     */
    public static final void bindResource(String key, Object resource) {
        
        Map<String, Object> m = threadLocal.get();
        if (m==null) {
            m=new HashMap<String, Object>();
            threadLocal.set(m);
        }
        if (log.isDebugEnabled())
            log.debug("Binding resource ["+resource+"] for factory ["+key+"] to current thread ["+Thread.currentThread().getName()+"]");
        m.put(key,resource);
    }
    
    /**
     * Unbind a resource from a singleton factory from the current thread, which has previously been
     * bound using {@link #bindResource(Object, Object)}.
     * 
     * If no resource has been bound to the current thread before, <code>null</code> is returned
     * and a warning is logged.
     * 
     * @param key A key for storing the given resource, usually a unique
     *            string like a JDBC URL or a remoting connection address.
     * @return The previously bound resource like a JDBC connection or a remoting context. 
     */
    public static final <T> T unbindResource(String key) {
        
        Map<String, Object> m = threadLocal.get();
        
        Object resource = null;
        
        if (m!=null)
            resource = m.remove(key);

        if (resource == null && log.isWarnEnabled())
            log.warn("Try to unbind missing resource for factory ["+key+"] from current thread ["+Thread.currentThread().getName()+"]");

        if (resource != null && log.isDebugEnabled())
            log.debug("Unbinding resource ["+resource+"] for factory ["+key+"] from current thread ["+Thread.currentThread().getName()+"]");
        
        return (T) resource;
    }
    
    /**
     * Fetch a resource from a singleton factory from the current thread, which has previously been
     * bound using {@link #bindResource(Object, Object)}.
     * 
     * @param key A key for storing the given resource, usually a unique
     *            string like a JDBC URL or a remoting connection address.
     * @return The previously bound resource like a JDBC connection or a remoting context.
     *         <code>null</code> is returned, if no resource has been bound to the current
     *         thread  under the given key.
     */
    public static final <T> T getBoundResource(String key) {
        
        Map<String, Object> m = threadLocal.get();
        
        Object resource = null;
        if (m!=null) resource = m.get(key);
        
        if (log.isTraceEnabled())
            log.trace("Fetching resource ["+resource+"] for factory ["+key+"] from current thread ["+Thread.currentThread().getName()+"]");

        return (T) resource;
    }
}
