/***********************************************************
 * $Id: GenericDAO.java 418 2011-01-01 22:28:15Z util $
 * 
 * JPA Utility classes of the clazzes.org project
 * http://www.clazzes.org
 *
 * Created: 2006-03-25
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 ***********************************************************/

package org.clazzes.util.aop.generic;

import java.io.Serializable;

/**
 *  This interface is implemented by DAOs, which support the traditional
 *  save, update and delete operations on a single data class.
 */
public interface GenericDAO<T>
{
    /**
     * @return The class object for the persistent class <code>T</code>. 
     */
    Class<T> getPersistentClass();
    
    /**
     * Returns the database object of the generic's type with the given
     * database Id.
     * 
     * @param id The database Id of the object to be retrieved.
     * @return The found object of type <tt>T</tt>.
     */
    T get(Serializable id);

    /**
     * Saves the given object of the generic's type to the database.
     * If the object has no database Id yet, a new instance of the object
     * will be returned, which carries a newly generated database Id.
     * 
     * @param t The object to be save to the database.
     * @return A copy of the created object of type <tt>T</tt> possibly
     *         with a freshly generated database Id.
     */
    T insert(T t);

    /**
     * Updates the given object of the generic's type in the database.
     * 
     * @param t The object to be updated in the database.
     * @return A copy of the object <tt>t</tt>.
     */
    T update(T t);

    /**
     * Merges (updates) the given object of the generic's type in the database.
     * 
     * @param t The object to be updated in the database.
     * @return A copy of the object <tt>t</tt>.
     */
    T merge(T t);


    /**
     * Refreshes (reloads, reverts) the given object of the generic's type from the database.
     * 
     * @param t The object to be reloaded from the database.
     * @return A copy of the object <tt>t</tt>.
     */
    T refresh(T t);

    /**
     * Deletes the given object of the generic's type from the database.
     * 
     * @param id Database-Id of object to be deleted from the database.
     */
    void delete(Serializable id);
    
}
