/***********************************************************
 * $Id: JdbcDAOSupport.java 402 2011-01-01 09:31:17Z util $
 * 
 * AOP Utility classes of the clazzes.org project
 * http://www.clazzes.org
 *
 * Created: 01.09.2010
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 ***********************************************************/

package org.clazzes.util.aop.jdbc;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Statement;

import org.clazzes.util.aop.DAOException;
import org.clazzes.util.aop.ThreadLocalManager;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A base class for JDBC DAO classes, which provides access to a thread-bound
 * JDBC connection, which is established by a {@link JdbcTransactionInterceptor}.
 */
public class JdbcDAOSupport
{
    private static final Logger log = LoggerFactory.getLogger(JdbcDAOSupport.class);
    
    private String threadLocalKey; 
    
    protected JdbcDAOSupport() {
        this.threadLocalKey = JdbcTransactionInterceptor.DEFAULT_THREAD_LOCAL_KEY;
    }
    
    /**
     * <p>Fetch previously bound connection from {@link ThreadLocalManager}.</p>
     * 
     * <p>Please carefully call {@link #closeStatement(Statement)} in a finally clause 
     * on statements you get from the retrieved connection or you will run in severe
     * resource leakages and threading problems.</p>
     * 
     * <p>It is recommended to use
     * {@link #performWithPreparedStatement(String, JdbcPreparedStatementAction)} and
     * {@link #performWithStatement(JdbcStatementAction)} whenever possible
     * in order to circumvent these problems.</p>
     * 
     * <p>If you set up JDBC statements on your own, please follow the pattern outlined below
     * in your implementation:</p>
     * <pre>
     * public RetValType myDAOFunction() {
     * 
     *   Connection connection = this.getConnection();
     *   Statement statement1 = null;
     *   PreparedStatement statement2 = null;
     *   try
     *   {
     *       statement1 = connection.createStatement();
     *       statement2 = connection.prepareStatement("select foo from bar where id = ?");
     *       
     *       ...your code operating with the two statements...
     *       
     *       return __your_result_here__;
     *       
     *   } catch (Exception e)
     *   {
     *       throw new DAOException(e);
     *   }
     *   finally {
     *       closeStatement(statement1);
     *       closeStatement(statement2);
     *   }
     * }  
     * </pre>
     * 
     * @return A JDBC transaction, which has previously been bound to the current thread
     *         using the configured thread-local key to
     *         {@link ThreadLocalManager#bindResource(Object, Object)}.
     */
    protected Connection getConnection() {
        
        Connection ret = ThreadLocalManager.getBoundResource(this.threadLocalKey);
        
        if (ret == null)
            throw new DAOException("No connection bound under key ["+this.threadLocalKey+"], function called outside a transaction context.");
        
        return ret;
    }

    /**
     * Close an SQL statement. Any exception thrown will be logged an
     * this function silently returns.
     * 
     * Call this function from a finally clause inside your DAO implementation,
     * if you do not use
     * {@link #performWithPreparedStatement(String, JdbcPreparedStatementAction)} or
     * {@link #performWithStatement(JdbcStatementAction)}.
     * 
     * @param statement The statement to close. If <code>null</code> is passed,
     *                  this method silently exits.
     */
    protected static void closeStatement(Statement statement) {
        if (statement == null) return;
        try
        {
            statement.close();
        } catch (SQLException e)
        {
            log.warn("Caught an exception while closing a JDBC statement",e);
        }
    }
    
    /**
     * Perform a JDBC action inside a context, where an JDBC statement has been set up
     * and will be closed in a finally clause.
     * 
     * @param <T> The intended return value.
     * @param action The action to perform on a newly created statement.
     * @return The value returned by {@link JdbcStatementAction#perform(Statement)}.
     */
    protected <T> T performWithStatement(JdbcStatementAction<T> action) {
        
        Connection connection = this.getConnection();
        Statement statement = null;
        try
        {
            statement = connection.createStatement();
            
            return action.perform(statement);
            
        } catch (Exception e)
        {
            throw new DAOException(e);
        }
        finally {
            closeStatement(statement);
        }
    }
    
    /**
     * Perform a JDBC action inside a context, where an JDBC statement has been set up
     * and will be closed in a finally clause.
     * 
     * @param <T> The intended return value.
     * @param sql The SQL code for the prepared statement.
     * @param action The action to perform on a newly created prepared statement based
     *               on the given SQL code. 
     * @return The value returned by {@link JdbcStatementAction#perform(Statement)}.
     */
    protected <T> T performWithPreparedStatement(String sql, JdbcPreparedStatementAction<T> action) {
        
        Connection connection = this.getConnection();
        PreparedStatement statement = null;
        try
        {
            statement = connection.prepareStatement(sql);
            
            return action.perform(statement);
            
        } catch (Exception e)
        {
            throw new DAOException(e);
        }
        finally {
            closeStatement(statement);
        }
    }
    
    /**
     * @return The key for retrieving the thread-bound connection.
     *         The default value is {@link #DEFAULT_THREAD_LOCAL_KEY}.
     */
    public String getThreadLocalKey()
    {
        return this.threadLocalKey;
    }

    /**
     * @param key the key for retrieving the connection to set.
     *            The default value is {@link #DEFAULT_THREAD_LOCAL_KEY}.
     *            Configure this property, if you use multiple JDBC datasources.
     *            For clarity, it is recommended to use the JDBC URL of the underlying
     *            datasource for this value.
     */
    public void setThreadLocalKey(String key)
    {
        this.threadLocalKey = key;
    }
}
