/***********************************************************
 * $Id: JdbcTransactionInterceptor.java 402 2011-01-01 09:31:17Z util $
 * 
 * AOP Utility classes of the clazzes.org project
 * http://www.clazzes.org
 *
 * Created: 31.08.2010
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 ***********************************************************/

package org.clazzes.util.aop.jdbc;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.SQLException;

import javax.sql.DataSource;

import org.aopalliance.intercept.MethodInterceptor;
import org.aopalliance.intercept.MethodInvocation;
import org.clazzes.util.aop.DAOException;
import org.clazzes.util.aop.ThreadLocalManager;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * <p>This method interceptor fetches a JDBC transaction from a JDBC
 * Datasource, opens a transaction and binds the JDBC connection to the current thread
 * using {@link ThreadLocalManager#bindResource(Object, Object)} with the given
 * datasource as a key.</p>
 * 
 * <p>When the method finishes, the transaction is committed or rolled back and
 * the JDBC connection is unbound from the current thread.</p>
 */
public class JdbcTransactionInterceptor implements MethodInterceptor
{
    private static final Logger log = LoggerFactory.getLogger(JdbcTransactionInterceptor.class);
    
    private DataSource dataSource;
    private Integer transactionIsolation;
    private String catalog;
    private String threadLocalKey;
    private boolean pedanticProbe;

    /**
     * The default thread local key, which is used if no other key is configured.
     */
    public static final String DEFAULT_THREAD_LOCAL_KEY = "org.clazzes.util.aop.jdbcConnection.0";
    
    public JdbcTransactionInterceptor() {
        this.threadLocalKey = DEFAULT_THREAD_LOCAL_KEY;
        this.pedanticProbe = true;
    }
    
    /**
     * This method may be used as an initialization method in order to resolve the databse driver and
     * check the database connection. If {@link #isPedanticProbe()} is <code>true</code>,
     * this method will fail with a {@link DAOException}, when the database connection
     * is not ready.
     */
    public void probeDataBase() {
        
        Connection connection = null;
            
        try
        {
            log.info("Probing database of datasource ["+this.dataSource+"]...");

            connection = this.dataSource.getConnection();
            
            DatabaseMetaData md = connection.getMetaData();
            
            log.info("Connected to ["+md.getDatabaseProductName()+"], version ["+md.getDatabaseProductVersion()+"] using driver ["+md.getDriverName()+"], version ["+md.getDriverVersion()+"].");
            
        } catch (SQLException e)
        {
            if (this.pedanticProbe) {
                log.error("Error probing database of datasource ["+this.dataSource+"]",e);
                throw new DAOException("Error probing database of datasource ["+this.dataSource+"]",e);
            }
            else
                log.warn("Error probing database of datasource ["+this.dataSource+"], maxbe you should start your database.",e);
        }
        finally {
            try
            {
                if (connection != null)
                    connection.close();
            } catch (SQLException e)
            {
                log.warn("Error closing probe connection of datasource ["+this.dataSource+"]",e);
            }
        }
    }
    
    /* (non-Javadoc)
     * @see org.aopalliance.intercept.MethodInterceptor#invoke(org.aopalliance.intercept.MethodInvocation)
     */
    @Override
    public Object invoke(MethodInvocation invocation) throws Throwable
    {
        Connection connection = this.dataSource.getConnection();
        
        if (log.isDebugEnabled())
            log.debug("Opening transaction for datasource ["+this.dataSource+"].");

        try {
            connection.setAutoCommit(false);
            if (this.transactionIsolation != null)
                connection.setTransactionIsolation(this.transactionIsolation.intValue());
            if (this.catalog != null)
                connection.setCatalog(this.catalog);
        } catch (Throwable e) {
            
            try {
                connection.close();
            } catch (SQLException sqle) {
                if (log.isWarnEnabled())
                    log.warn("Exception caught closing JDBC transaction upon connection setup error.",sqle);
            }
            throw e;
        }
        
        ThreadLocalManager.bindResource(this.threadLocalKey,connection);
        
        try {
            Object ret = invocation.proceed();
            connection.commit();
            return ret;
            
        } catch (Throwable e) {
        
            try {
                if (log.isDebugEnabled())
                    log.debug("Rolling back transaction for datasource ["+this.dataSource+"].");
                connection.rollback();
                
            } catch (SQLException sqle) {
                if (log.isWarnEnabled())
                    log.warn("Exception caught while rolling back JDBC transaction",sqle);
            }
            throw e;
        }
        finally {
            ThreadLocalManager.unbindResource(this.threadLocalKey);
            try {
                if (log.isDebugEnabled())
                    log.debug("Closing connection for datasource ["+this.dataSource+"].");
                connection.close();
            } catch (SQLException sqle) {
                if (log.isWarnEnabled())
                    log.warn("Exception caught closing JDBC transaction",sqle);
            }
        }
    }

    /**
     * @return The JDBC datasource for retrieving the connection.
     */
    public DataSource getDataSource()
    {
        return this.dataSource;
    }

    /**
     * @param dataSource the JDBC datasource for retrieving the connection to set.
     */
    public void setDataSource(DataSource dataSource)
    {
        this.dataSource = dataSource;
    }

    /**
     * @return The key for retrieving the thread-bound connection.
     *         The default value is {@link #DEFAULT_THREAD_LOCAL_KEY}.
     */
    public String getThreadLocalKey()
    {
        return this.threadLocalKey;
    }

    /**
     * @param key the key for retrieving the connection to set.
     *            The default value is {@link #DEFAULT_THREAD_LOCAL_KEY}.
     *            Configure this property, if you use multiple JDBC datasources.
     *            For clarity, it is recommended to use the JDBC URL of the underlying
     *            datasource for this value.
     */
    public void setThreadLocalKey(String key)
    {
        this.threadLocalKey = key;
    }

    /**
     * @return An optional transaction isolation level to set.
     * @see Connection#setTransactionIsolation(int)
     */
    public Integer getTransactionIsolation()
    {
        return this.transactionIsolation;
    }

    /**
     * @param transactionIsolation the optional transaction isloation level to set.
     */
    public void setTransactionIsolation(Integer transactionIsolation)
    {
        this.transactionIsolation = transactionIsolation;
    }

    /**
     * @return An optional catalog to set to newly created connection.
     * @see Connection#setCatalog(String)
     */
    public String getCatalog()
    {
        return this.catalog;
    }

    /**
     * @param catalog the optional catalog to set to newly created connection to set
     */
    public void setCatalog(String catalog)
    {
        this.catalog = catalog;
    }

    /**
     * @return Whether {@link #probeDataBase()} should fail on connection errors.
     */
    public boolean isPedanticProbe()
    {
        return this.pedanticProbe;
    }

    /**
     * @param pedanticProbe Whether {@link #probeDataBase()} should fail on connection errors.
     */
    public void setPedanticProbe(boolean pedanticProbe)
    {
        this.pedanticProbe = pedanticProbe;
    }

}
